/*
 * host.c - host for multiap network ubus interface
 *
 * Copyright (C) 2021 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdbool.h>
#include <pthread.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <uci.h>

#include <easy/easy.h>

#include <netlink/netlink.h>

#include "debug.h"
#include "host_config.h"
#include "topologyd.h"
#include "host_nodes.h"
#include "config.h"
#include "json_utils.h"
#include "host_utils.h"
#include "host.h"
#include "ieee1905_deps.h"
#include "host_ieee1905_utils.h"
#include "host_wan_stats.h"

extern const char *ubus_socket;

struct u_list {
	char ifname[50][128];
	uint32_t count;
};

void remove_newline(char *buf)
{
	int len;

	len = strlen(buf) - 1;
	if (buf[len] == '\n')
		buf[len] = 0;
}

char *topology_stringify_address_types(enum address_type type)
{
	switch (type) {
	case HOST_ADDR_STATIC:
		return "Static";
	case HOST_ADDR_DHCP:
		return "DHCP";
	}
	return "";
}

char *topology_stringify_types(enum host_types type)
{
	switch (type) {
	case HOST_TYPE_ETHER:
		return "Ethernet";
	case HOST_TYPE_GHN:
		return "G.hn";
	case HOST_TYPE_HPNA:
		return "HPNA";
	case HOST_TYPE_HOMEPLUG:
		return "HomePlug";
	case HOST_TYPE_MOCA:
		return "MoCA";
	case HOST_TYPE_UPA:
		return "UPA";
	case HOST_TYPE_WIFI:
		return "Wi-Fi";
	case HOST_TYPE_OTHER:
		return "Other";
	}
	return "";
}

static void uobjx_lookup(struct ubus_context *ctx, struct ubus_object_data *obj, void *priv)
{
	const char *token = "wifi.ap";
	char *p;
	struct u_list *str = (struct u_list *)priv;

	p = strstr(obj->path, token);
	if (p != NULL) {
		strncpy(str->ifname[str->count], obj->path, 16);
		str->count = str->count + 1;
	}
}

static int host_get_wifi_nodes(struct ubus_context *ctx, struct u_list *c)
{
	c->count = 0;
	int i = 0;

	ubus_lookup(ctx, NULL, uobjx_lookup, c);
	for (i = 0; i < c->count; i++)
		dbg("%d %s\n", __LINE__, c->ifname[i]);
	return 0;
}

int host_nodes(struct ubus_context *ctx,
		struct ubus_object *obj,
		struct ubus_request_data *req,
		const char *method,
		struct blob_attr *msg)
{
	struct blob_buf bb;
	struct topologyd_private *priv =
		container_of(obj, struct topologyd_private, obj);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	host_dump_node_nbr(priv, &bb);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

void host_dump_node_nbr(struct topologyd_private *priv, struct blob_buf *b)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	struct uci_section *section;
	char mac_str[18] = { 0x0 };
	void *nodes_array;
	struct host_node *p;
	void *table1 = NULL;
	struct tm *info;
	time_t tmp_t;
	char str_tm[32] = {0};
	int j = 0;

	ctx = uci_alloc_context();
	if (!ctx)
		return;

	if (priv->host_global_cfg.persistent == 0) {
		uci_set_confdir(ctx, "/var/state");
	}

	if (uci_load(ctx, "hosts", &pkg)) {
		uci_free_context(ctx);
		return;
	}

	dbg("Inside %s %d\n", __func__, __LINE__);
	nodes_array = blobmsg_open_array(b, "hosts");
	uci_foreach_element(&pkg->sections, e) {
		const char *macstr;
		uint8_t hwaddr[6] = {0};
		void *arr, *table;

		section = uci_to_section(e);
		if (strncmp(section->type, "host", 4))
			continue;

		macstr = uci_lookup_option_string(ctx, section, "macaddr");
		if (!macstr)
			continue;

		hwaddr_aton(macstr, hwaddr);

		p = host_node_lookup(priv->host.node_htable, hwaddr);
		if (p && (p->ignore != 1)) {
			char linkstr[18] = {0};
			struct host_node *al_node = p->al_node;

			if (al_node && memcmp(al_node->link_hwaddr, al_node->hwaddr, 6)) {
				if (!hwaddr_ntoa(al_node->link_hwaddr, linkstr)) {
					err("|%s:%d| invalid link hwaddr for node %s\n",
					    __func__, __LINE__, macstr);
					continue;
				}
			} else {
				if (!hwaddr_ntoa(p->link_hwaddr, linkstr)) {
					err("|%s:%d| invalid link hwaddr for node %s\n",
					    __func__, __LINE__, macstr);
					continue;
				}
			}
			table1 = blobmsg_open_table(b, NULL);
			hwaddr_ntoa(p->hwaddr, mac_str);
			blobmsg_add_string(b, "macaddr", macstr);
			blobmsg_add_string(b, "link_macaddr", linkstr);

			blobmsg_add_string(b, "ipaddr", p->ipv4addr);
			blobmsg_add_string(b, "address_source", topology_stringify_address_types(p->addrsrc));
			blobmsg_add_u64(b, "lease_time_remaining", p->leasetmrmn);
			blobmsg_add_string(b, "device", p->device);
			blobmsg_add_string(b, "parent_device", p->parent_device);
			blobmsg_add_string(b, "network", p->network);

			blobmsg_add_string(b, "interface_type", topology_stringify_types(p->intf_type));
			blobmsg_add_string(b, "hostname", p->hostname);

			blobmsg_add_u8(b, "active", p->active);
			tmp_t = p->active_last_change;
			info = gmtime(&tmp_t);
			strftime(str_tm, sizeof(str_tm), "%Y-%m-%dT%H:%M:%SZ", info);
			blobmsg_add_string(b, "active_last_change", str_tm);
			blobmsg_add_u64(b, "active_connections", p->active_connections);
			//Here display the list of IPV4 addresses

			arr = blobmsg_open_array(b, "ipv4addr");
			for (j = 0; j < p->ipv4addr_count; j++)
				blobmsg_add_string(b, NULL, p->ipv4addr_list[j]);
			blobmsg_close_array(b, arr);
			//Here we display the list of IPV6 addresses
			arr = blobmsg_open_array(b, "ipv6addr");
			for (j = 0; j < p->ipv6addr_count; j++)
				blobmsg_add_string(b, NULL, p->ipv6addr[j]);
			blobmsg_close_array(b, arr);
			table = blobmsg_open_table(b, "stats");
			blobmsg_add_u64(b, "tx_packets", p->stats.tx_packets);
			blobmsg_add_u64(b, "tx_bytes", p->stats.tx_bytes);
			blobmsg_add_u64(b, "rx_packets", p->stats.rx_packets);
			blobmsg_add_u64(b, "rx_bytes", p->stats.rx_bytes);
			blobmsg_close_table(b, table);
			blobmsg_close_table(b, table1);
		}
	}

	blobmsg_close_array(b, nodes_array);

	uci_unload(ctx, pkg);
	uci_free_context(ctx);
}

struct host_node *host_check_get_node(struct topologyd_private *priv, uint8_t *mac, bool *send_ev)
{
	char mac_str[18] = { 0 };
	struct host_node *p = NULL;

	if (!memcmp(mac, priv->ieee1905_macaddr, 6))
		return NULL;

	if (!memcmp(mac, priv->albridge_macaddr, 6))
		return NULL;

	/* Add node if not present in host.
	 *
	 * In case the node is present, update the node's values
	 * and the timestamp.
	 */

	p = host_node_lookup(priv->host.node_htable, mac);
	if (!p) {
		p = host_node_add(priv->host.node_htable, mac);
		if (p == NULL)
			return NULL;
		*send_ev = true;
	}

	dbg("%s %d ||||||||| setting node "MACFMT" to active\n", __func__, __LINE__, MAC2STR(p->hwaddr));

	//here we need to write  this info in the config file
	hwaddr_ntoa(mac, mac_str);
	config_add_default_host_mac("hosts", "host", mac_str, priv->host_global_cfg.persistent);
	return p;

}

struct host_node *host_topo_node_add(struct topologyd_private *priv,
				     struct node *node, uint8_t *mac_addr,
				     uint8_t is_wifi)
{
	struct host_node *p = NULL;
	char mac_str[18] = { 0 };
	int ret = 0;
	bool send_ev = false;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!priv)
		return NULL;

	if (priv->host.num_nodes >= HOST_NODE_MAX) {
		err("Number of host more than MAX limit!\n");
		return NULL;
	}

	if (node != NULL) {
		p = host_check_get_node(priv, node->hwaddr, &send_ev);
		if (p == NULL)
			return NULL;
		p->is_ieee1905_node = 1;

		if (p->intf_type != HOST_TYPE_WIFI || is_wifi)
			host_change_active_state(priv, p, 1);

		host_get_wifi_data(priv, p);

	} else if (mac_addr != NULL) {
		p = host_check_get_node(priv, mac_addr, &send_ev);
		if (p == NULL)
			return NULL;
		hwaddr_ntoa(mac_addr, mac_str);
		p->is_ieee1905_node = 0;
		if (p->active != 1) {
			dbg("%s %d ||||||||| setting node %s to active\n", __func__, __LINE__, mac_str);
			if (p->intf_type != HOST_TYPE_WIFI || is_wifi)
				host_change_active_state(priv, p, 1);
		}

		if (is_wifi == 0 && p->intf_type != HOST_TYPE_WIFI) {
			ret = host_get_wifi_data(priv, p);
			if (ret != 0)
				err("Unable to fetch interface type\n");
		} else {
			config_set_host_option("hosts", "host", "macaddr",
						mac_str, "interface_type",
						"wifi",
						priv->host_global_cfg.persistent);
			p->intf_type = HOST_TYPE_WIFI;
		}

		config_set_host_option("hosts", "host", "macaddr", mac_str,
					"interface_type",
					(p->intf_type == HOST_TYPE_WIFI ? "wifi" : "eth"),
					priv->host_global_cfg.persistent);
	}

	//bool has_ip = p->is_ipaddr;

	ret = host_get_ipaddr_hostname(priv, mac_str, p);
	if (ret != 0) {
		err("Unable to fetch ip adddress and host name\n");
		p->is_ipaddr = 0;
	} else {
		p->is_ipaddr = 1;
	}

	if (p->intf_type == HOST_TYPE_WIFI)
		p->age_time = time(NULL);


	if (p && p->al_node)
		host_topo_node_add(priv, NULL, p->al_node->hwaddr, is_wifi);

	return p;
}

struct host_node *host_topo_node_del(struct topologyd_private *priv,
				     struct node *node, uint8_t *mac_addr,
				     uint8_t is_wifi)
{
	struct host_node *p = NULL;
	char mac_str[18] = { 0 };

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!priv)
		return NULL;

	if (node != NULL) {
		if (!memcmp(node->hwaddr, priv->ieee1905_macaddr, 6)) {
			//self node does not get deleted in the hosts
			return NULL;
		}

		/*
		 * In case the node is present, update the node's values
		 * and the timestamp.
		 */
		p = host_node_lookup(priv->host.node_htable, node->hwaddr);
		if (p) {
			if (p->active)
				host_send_client_event(priv, p, 0);
			host_change_active_state(priv, p, 0);
			p->is_ipaddr = 0;

		} else {
			return NULL;
		}
	} else if (mac_addr != NULL) {
		p = host_node_lookup(priv->host.node_htable, mac_addr);
		if (p) {
			if (p->active != 0) {
				if (p->is_ipaddr)
					host_send_client_event(priv, p, 0);
				host_change_active_state(priv, p, 0);
			}
		} else {
			return NULL;
		}
	}
	if (p && is_wifi) {
		config_set_host_option("hosts", "host", "macaddr", mac_str,
					"interface_type",
					"wifi",
					priv->host_global_cfg.persistent);
		p->intf_type = HOST_TYPE_WIFI;
	}

	if (p && p->al_node)
		host_topo_node_del(priv, NULL, p->al_node->hwaddr, is_wifi);

	if (p) {
		p->is_event = 0;
	}

	return p;
}

//This function is used to check if the host is present in the  dhcpleases*/
int host_check_dhcpv4_info(uint8_t *mac_list, int intf_num, uint8_t *mac)
{
	FILE *leases;
	char line[256];
	long leasetime, tmplease = -1;
	char macaddr[24];
	char ipaddr[24];
	char hostname[64];
	char clid[256];
	char mac_addr[18] = { 0 };
	int i = 0;


	leases = fopen("/var/dhcp.leases", "r");
	if (leases != NULL) {
		while (fgets(line, sizeof(line), leases) != NULL) {
			remove_newline(line);
			if (sscanf(line, "%ld %23s %23s %63s %255s", &leasetime, macaddr, ipaddr, hostname, clid) == 5) {
				dbg("%s %d ipv4addr = [%s]\n", __func__, __LINE__, ipaddr);
				for (i = 0; i < intf_num; i++) {
					hwaddr_ntoa(&(mac_list[i*6]), mac_addr);
					if (strcasecmp(mac_addr, macaddr) == 0) {
						if (leasetime > tmplease) {
							tmplease = leasetime;
							memcpy(mac, &(mac_list[i*6]), 6);
						}
					}
				}
			}
		}
		fclose(leases);
	}
	if (tmplease == -1)
		return 0;
	else
		return 1;
}



int host_append_dhcpv4_info(char *mac_addr, struct host_node *p)
{
	FILE *leases;
	char line[256];
	long leasetime;
	char macaddr[24];
	char ipaddr[24];
	char hostname[64];
	char clid[256];
	int found = 0;
	int ret = 0;

	dbg("Inside %s %d macis %s\n", __func__, __LINE__, mac_addr);

	if (!p)
		return -1;

	p->ipv4addr_count = 0;
	leases = fopen("/var/dhcp.leases", "r");
	if (leases != NULL) {
		while (fgets(line, sizeof(line), leases) != NULL) {
			remove_newline(line);
			if (sscanf(line, "%ld %23s %23s %63s %255s", &leasetime, macaddr, ipaddr, hostname, clid) == 5) {
				dbg("%s %d ipv4addr = [%s]\n", __func__, __LINE__, ipaddr);
				if (strcasecmp(mac_addr, macaddr) == 0) {
					if (found == 0) {
						strncpy(p->hostname, hostname, 64);
						strncpy(p->ipv4addr, ipaddr, 24);
						found = 1;
						p->leasetmrmn = leasetime - (int)time(NULL);

					}
					ret = strncmp(ipaddr, "", 24);
					if (ret != 0) {
						strncpy(p->ipv4addr_list[p->ipv4addr_count], ipaddr, 24);
						p->ipv4addr_count = (p->ipv4addr_count + 1);
					}
				}
			}
		}
		fclose(leases);
	}
	if (found == 0)
		return 1;
	else
		return 0;
}

void host_append_dhcpv6_info(struct host_node *p)
{
	FILE *hosts6;
	char line[512];
	char hostname[64];
	char ip6addr[128];

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!p)
		return;

	p->ipv6addr_count = 0;
	hosts6 = fopen("/tmp/hosts/odhcpd", "r");
	if (hosts6 != NULL) {
		while (fgets(line, sizeof(line), hosts6) != NULL) {
			remove_newline(line);

			if (sscanf(line, "%127s %63s", ip6addr, hostname)) {
				if (strcasecmp(p->hostname, hostname) == 0) {
					dbg("ipv6 address %s %d\n", __func__, __LINE__);
					strncpy(p->ipv6addr[p->ipv6addr_count], ip6addr, 128);
					p->ipv6addr_count = p->ipv6addr_count + 1;
				}
			}
		}
		fclose(hosts6);
	}
}

void mask_num_to_bitmap(int num, uint32_t *map)
{
	int i;

	for (i = 0; i < num; i++)
		*map |= (1 << i);
}

void network_dump_parser(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct topologyd_private *p = (struct topologyd_private *)req->priv;
	struct json_object *json_msg;
	char *json_str;

	json_str = blobmsg_format_json(msg, true);
	if (!json_str)
		return;

	json_msg = json_tokener_parse(json_str);
	if (!json_msg)
		goto out_str;

	if (p->network_dump) {
		json_object_put(p->network_dump);
		p->network_dump = NULL;
	}

	if (!json_object_is_type(json_msg, json_type_object))
		goto out_json;

	p->network_dump = json_object_get(json_msg);

out_json:
	json_object_put(json_msg);
out_str:
	free(json_str);
}

void host_get_network(struct topologyd_private *priv, struct host_node *p)
{
	struct json_object *interfaces = NULL;
	int i = 0, len = 0;
	bool ret = 0;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!priv || !p)
		return;

	if (!json_object_is_type(priv->network_dump, json_type_object))
		return;

	ret = json_object_object_get_ex(priv->network_dump, "interface", &interfaces);
	if (!ret || !interfaces)
		return;

	if (!json_object_is_type(interfaces, json_type_array))
		return;

	len = json_object_array_length(interfaces);

	for (i = 0; i < len; i++) {
		struct json_object *sta_obj = NULL, *jobj_network = NULL, *ipv4_arr = NULL;
		const char *network;
		int ipv4_len, j;

		sta_obj = json_object_array_get_idx(interfaces, i);
		ret = json_object_object_get_ex(sta_obj, "interface", &jobj_network);
		if (!ret || !jobj_network)
			continue;

		network = json_object_get_string(jobj_network);
		if (!network)
			continue;

		dbg("found network = %s\n", network);
		ret = json_object_object_get_ex(sta_obj, "ipv4-address", &ipv4_arr);
		if (!ret || !ipv4_arr)
			continue;

		if (!json_object_is_type(ipv4_arr, json_type_array))
			return;

		ipv4_len = json_object_array_length(ipv4_arr);
		for (j = 0; j < ipv4_len; j++) {
			struct in_addr ip, snet, host, rslt;
			struct json_object *jobj_ptr = NULL, *iface = NULL;
			const char *ipaddr;
			int netmask;
			uint32_t mask = 0;

			iface = json_object_array_get_idx(ipv4_arr, j);
			if (!iface)
				continue;
			ret = json_object_object_get_ex(iface, "mask", &jobj_ptr);
			if (!ret || !jobj_ptr)
				continue;

			netmask = json_object_get_int(jobj_ptr);

			ret = json_object_object_get_ex(iface, "address", &jobj_ptr);
			if (!ret || !jobj_ptr)
				continue;

			ipaddr = json_object_get_string(jobj_ptr);
			if (!ipaddr)
				continue;

			mask_num_to_bitmap(netmask, &mask);

			inet_pton(AF_INET, ipaddr, &(ip.s_addr));
			inet_pton(AF_INET, p->ipv4addr, &(host.s_addr));

			snet.s_addr = (ip.s_addr & mask);
			rslt.s_addr = (host.s_addr & mask);
			if ((snet.s_addr ^ rslt.s_addr) == 0) {
				dbg("SUBNET %s MATCHED %s!!!\n", network, p->ipv4addr);
				snprintf(p->network, 16, "%s", network);
				return;
			}
		}
	}
}

void host_get_network_dump(struct topologyd_private *priv)
{
	struct blob_buf bb = {0};
	int ret;
	uint32_t id;
	int status;

	blob_buf_init(&bb, 0);

	dbg("Calling 'interfacedump'!!\n");

	if (!priv) {
		blob_buf_free(&bb);
		return;
	}

	status = ubus_lookup_id(priv->ctx, "network.interface", &id);
	if (status != UBUS_STATUS_OK) {
		err("object 'interfacedump' not present!\n");
		blob_buf_free(&bb);
		return;
	}

	ret = ubus_invoke(priv->ctx, id, "dump", bb.head,
			network_dump_parser, priv, 2 * 1000);
	if (ret) {
		dbg("Failed to get network.interface dump (ret = %d)\n", ret);
	}

	blob_buf_free(&bb);
}

void get_bridge_info(struct host_node *host, char *device)
{
	char brinfo[16];
	int aging, timer;
	char macstr[18] = {0};

	if (!host)
		return;

	hwaddr_ntoa(host->hwaddr, macstr);

	aging = 255;
	host->port_no = 0;
	if (strstr(device, "br-")) {
		chrCmd(brinfo, 16, "brctl showmacs %s | grep %s | awk '{print$1\":\"$4}'", device, macstr);
		sscanf(brinfo, "%d:%d.%d", &host->port_no, &aging, &timer);
	}
}


static void get_ethernet_interface(struct host_node *host, char *bridge, int portno)
{
	DIR *dir;
	struct dirent *ent;
	FILE *pfile;
	char path_brif[256];
	char path_portno[1024];
	unsigned int pno = 0;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (portno == 0 || !host)
		return;

	snprintf(path_brif, 256, "/sys/class/net/%s/brif/", bridge);

	dir = opendir(path_brif);

	if (dir != NULL) {
		while ((ent = readdir(dir)) != NULL) {
			if (ent->d_name[0] == '.')
				continue;

			snprintf(path_portno, 1024, "/%s/%s/port_no", path_brif, ent->d_name);
			pfile = fopen(path_portno, "r");

			if (pfile != NULL) {
				fscanf(pfile, "0x%x", &pno);
				fclose(pfile);
			}

			if (pno != 0 && portno == pno) {
				strncpy(host->device, ent->d_name, 16);
				break;
			}
		}
		closedir(dir);
	}
}

int get_ifaddr(char *ifname, char *buf)
{
	int fd;
	struct ifreq ifr;
	int rv = -1;

	if (!ifname || !buf)
		return rv;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (!fd)
		return rv;

	ifr.ifr_addr.sa_family = AF_INET;

	strncpy(ifr.ifr_name, ifname, 15);
	rv = ioctl(fd, SIOCGIFADDR, &ifr);
	if (!rv)
		strncpy(buf, inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr), 15);

	close(fd);

	/* display result */

	return rv;
}

int get_ifmask(char *ifname, char *buf)
{
	int fd;
	struct ifreq ifr;
	int rv = -1;

	if (!ifname || !buf)
		return rv;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (!fd)
		return rv;

	ifr.ifr_addr.sa_family = AF_INET;

	strncpy(ifr.ifr_name, ifname, 15);
	rv = ioctl(fd, SIOCGIFNETMASK, &ifr);
	if (!rv)
		strncpy(buf, inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr), 15);

	close(fd);

	/* display result */

	return rv;
}

int host_get_parent_iface(struct topologyd_private *priv, struct host_node *p)
{
	struct json_object *iface_array;
	int i = 0, len = 0;

	if (!p || !priv || !priv->ieee1905_interfaces)
		return -1;

	if (!json_object_is_type(priv->ieee1905_interfaces, json_type_object))
		return -1;

	json_object_object_get_ex(priv->ieee1905_interfaces, "interface", &iface_array);
	if (!iface_array)
		return -1;

	if (!json_object_is_type(iface_array, json_type_array))
		return -1;

	len = json_object_array_length(iface_array);
	for (i = 0; i < len; i++) {
		struct json_object *iface_obj, *iface_obj_ifname, *iface_obj_parent;
		const char *ifname, *parent_ifname;

		iface_obj = json_object_array_get_idx(iface_array, i);
		if (!iface_obj)
			continue;

		json_object_object_get_ex(iface_obj, "ifname", &iface_obj_ifname);
		if (!iface_obj_ifname)
			continue;

		ifname = json_object_get_string(iface_obj_ifname);
		if (!ifname)
			continue;

		if (strncmp(ifname, p->device, sizeof(p->device)))
			continue;

		json_object_object_get_ex(iface_obj, "parent_ifname", &iface_obj_parent);
		if (!iface_obj_parent)
			continue;

		parent_ifname = json_object_get_string(iface_obj_parent);
		if (!parent_ifname)
			continue;

		strncpy(p->parent_device, parent_ifname, sizeof(p->parent_device));
		break;
	}

	return 0;
}

int host_get_ipaddr_hostname(struct topologyd_private *priv,
		char *macaddr_str, struct host_node *p)
{
	int ret = -1;
	char device[16] = { 0 };
	char ipv4_address[24] = { 0 };

	if (!macaddr_str || !p)
		return ret;

	ret = host_get_netlink_ip(p->hwaddr, ipv4_address, device);

	if (ret != 0)
		return ret;

	strncpy(p->ipv4addr, ipv4_address, 24);
	dbg("The IPv4 address is %s mac is %s device is %s\n", p->ipv4addr, macaddr_str, device);

	host_get_netlink_ip6(p->hwaddr, p);

	ret = host_append_dhcpv4_info(macaddr_str, p);
	if (ret != 0) {
		strncpy(p->ipv4addr_list[0], p->ipv4addr, 24);
		p->leasetmrmn = 0;
		p->addrsrc = HOST_ADDR_STATIC;
		p->ipv4addr_count = 1;
	} else {
		p->addrsrc = HOST_ADDR_DHCP;
	}
	host_get_network(priv, p);
	get_bridge_info(p, device);

	if ((p->intf_type == HOST_TYPE_WIFI && p->is_ieee1905_node) ||
	    (p->intf_type == HOST_TYPE_WIFI && !p->is_in_assoclist)) {
		strncpy(p->device, device, 16);
		get_ethernet_interface(p, device, p->port_no);
	} else if (p->intf_type == HOST_TYPE_ETHER) {
		strncpy(p->device, device, 16);
		get_ethernet_interface(p, device, p->port_no);
	}

	host_get_parent_iface(priv, p);

	ret = strncmp(p->ipv4addr, "", 24);
	if (ret == 0)
		return -1;

	dbg("%s %d ipv4addr = [%s]\n", __func__, __LINE__, p->ipv4addr);
	// Append DHCPv6 information: IPv6 Address, DUID, Lease Time
	//if (strlen(p->hostname) > 1)
		//host_append_dhcpv6_info(p);
	ret = 0;
	dbg("Inside %s %d mac=[%s] macinterface =[%d]\n", __func__, __LINE__, macaddr_str, p->intf_type);

	return ret;
}

bool host_is_in_assoclist(struct topologyd_private *priv, uint8_t *hwaddr)
{
	struct json_object *sta_array;
	const char *macstr;
	int i = 0, len = 0;
	int ret = 0;

	if (!hwaddr)
		return false;

	if (!json_object_is_type(priv->wifi_stations, json_type_object))
		return false;

	json_object_object_get_ex(priv->wifi_stations, "stations", &sta_array);
	if (!sta_array)
		return false;

	len = json_object_array_length(sta_array);

	for (i = 0; i < len; i++) {
		struct json_object *sta_obj, *sta_obj_macaddr;
		unsigned char macaddr[6] = {0};

		sta_obj = json_object_array_get_idx(sta_array, i);
		if (!sta_obj)
			continue;

		json_object_object_get_ex(sta_obj, "macaddr", &sta_obj_macaddr);
		if (!sta_obj_macaddr)
			continue;

		macstr = json_object_get_string(sta_obj_macaddr);
		if (!macstr)
			continue;

		hwaddr_aton(macstr, macaddr);
		ret = memcmp(macaddr, hwaddr, 6);
		if (ret)
			continue;

		dbg("%s: sta: is Wifi %s\n", __func__, macstr);
		return true;
	}

	return false;
}


int host_get_wifi_data(struct topologyd_private *priv, struct host_node *p)
{
	struct json_object *sta_array;
	const char *macstr;
	int i = 0, len = 0;
	int ret = 0;

	if (!p)
		return -1;

	p->is_in_assoclist = false;

	if (!json_object_is_type(priv->wifi_stations, json_type_object))
		return -1;

	json_object_object_get_ex(priv->wifi_stations, "stations", &sta_array);
	if (!sta_array)
		return -1;
	len = json_object_array_length(sta_array);

	for (i = 0; i < len; i++) {
		struct json_object *sta_obj, *sta_obj_macaddr, *jobj_ptr;
		unsigned char macaddr[6] = {0};

		sta_obj = json_object_array_get_idx(sta_array, i);
		if (!sta_obj)
			continue;

		json_object_object_get_ex(sta_obj, "macaddr", &sta_obj_macaddr);
		if (!sta_obj_macaddr)
			continue;

		macstr = json_object_get_string(sta_obj_macaddr);
		if (!macstr)
			continue;

		hwaddr_aton(macstr, macaddr);
		ret = memcmp(macaddr, p->hwaddr, 6);
		if (ret)
			continue;

		json_object_object_get_ex(sta_obj, "wdev", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		strncpy(p->device, json_object_get_string(jobj_ptr), 16);
		host_get_parent_iface(priv, p);
		json_object_object_get_ex(sta_obj, "frequency", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->frequency = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(sta_obj, "rssi", &jobj_ptr);
		if (!jobj_ptr)
			continue;
		p->rssi = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(sta_obj, "snr", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->snr = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(sta_obj, "idle", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->idle = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(sta_obj, "in_network", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->in_network = json_object_get_int64(jobj_ptr);

		dbg("%s: sta: is Wifi %s\n", __func__, macstr);
		p->intf_type = HOST_TYPE_WIFI;
		if (p->active != 1)
			host_change_active_state(priv, p, 1);
		p->is_in_assoclist = true;
		break;
	}

	return 0;
}

void host_gather_stations(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct topologyd_private *p = (struct topologyd_private *)req->priv;
	struct json_object *json_msg;
	struct json_object *sta_array;
	struct json_object *priv_stations;
	char *json_str;
	int i = 0, len = 0, priv_len = 0;

	json_object_object_get_ex(p->wifi_stations, "stations", &priv_stations);
	if (!priv_stations)
		return;

	priv_len = json_object_array_length(priv_stations);

	json_str = blobmsg_format_json(msg, true);
	if (!json_str)
		return;

	json_msg = json_tokener_parse(json_str);
	if (!json_msg)
		goto out_str;

	if (!json_object_is_type(json_msg, json_type_object))
		goto out_json;

	json_object_object_get_ex(json_msg, "stations", &sta_array);
	len = json_object_array_length(sta_array);

	for (i = 0; i < len; i++) {
		struct json_object *sta_obj;

		sta_obj = json_object_array_get_idx(sta_array, i);
		json_object_array_put_idx(priv_stations, priv_len++, json_object_get(sta_obj));
	}

out_json:
	json_object_put(json_msg);
out_str:
	free(json_str);
}

static int host_invoke_wifi_nodes(struct topologyd_private *priv,
		char *intf_name)
{
	struct blob_buf bb = {};
	int ret;
	uint32_t id;
	int status;

	blob_buf_init(&bb, 0);

	if (priv == NULL || intf_name == NULL)
		return -1;

	dbg("Calling 'assoclist' for fh-iface = %s\n", intf_name);

	status = ubus_lookup_id(priv->ctx, intf_name, &id);
	if (status != UBUS_STATUS_OK) {
		err("object '%s' not present!\n", intf_name);
		blob_buf_free(&bb);
		return OBJECT_INVALID;
	}

	ret = ubus_invoke(priv->ctx, id, "stations", bb.head,
			host_gather_stations, priv, 2 * 1000);
	if (ret) {
		dbg("Failed to get assoclist (ret = %d)\n", ret);
		blob_buf_free(&bb);
		return -1;
	}

	blob_buf_free(&bb);
	return 0;

}

int host_get_interface_type(struct topologyd_private *priv)
{
	struct json_object *arr;
	struct u_list c;
	int i = 0;
	int ret = 0;

	if (priv->wifi_stations) {
		json_object_put(priv->wifi_stations);
		priv->wifi_stations = NULL;
	}

	priv->wifi_stations = json_object_new_object();
	if (!priv->wifi_stations)
		return -1;

	arr = json_object_new_array();
	if (!arr)
		return -1;

	json_object_object_add(priv->wifi_stations, "stations", arr);
	host_get_wifi_nodes(priv->ctx, &c);
	for (i = 0 ; i < c.count ; i++) {
		ret = host_invoke_wifi_nodes(priv, c.ifname[i]);
		if (ret != 0) {
			err("Unable to get the assoclist\n");
			return -1;
		}
	}

	return ret;
}

int host_get_neigh_status(uint8_t *mac_addr)
{
	char cmd[256] = {0};
	FILE *fp = NULL;
	char *line = NULL;
	size_t len = 0;
	char macaddr_str[18];
	int found = 0;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (mac_addr == NULL)
		return -1;

	hwaddr_ntoa(mac_addr, macaddr_str);
	snprintf(cmd, 256,
		 "ip neigh show nud reachable|awk '{print $5}'|grep %s",
		 macaddr_str);

	fp = popen(cmd, "r");	/* flawfinder: ignore */
	if (fp == NULL) {
		err("unable to execute command");
		return -1;
	}

	dbg("Inside %s %d command is %s\n", __func__, __LINE__, cmd);
	while (getline(&line, &len, fp) != -1) {
		if (len < 17) {
			free(line);
			len = 0;
			continue;
		}
		found = 1;
		break;
	}

	free(line);
	pclose(fp);

	if (found == 0)
		return -1;
	else
		return 0;
}

void host_get_table_cb(struct topologyd_private *priv)
{
	struct host_node *p;
	int i = 0, ret = -1;
	char mac_str[18] = { 0 };

	dbg("Inside %s %d\n", __func__, __LINE__);
	if (!priv)
		return;

	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->host.node_htable[i]))
			continue;

		hlist_for_each_entry(p, &priv->host.node_htable[i], hlist) {
			bool has_ip = p->is_ipaddr;

			//if (p->al_node)
			//	continue;

			if (hwaddr_is_zero(p->hwaddr))  // TODO: unlikely..
				continue;

			if (p->is_copy == 1)
				continue;

			hwaddr_ntoa(p->hwaddr, mac_str);

			if (p->intf_type != HOST_TYPE_ETHER && !p->is_ieee1905_node && p->is_device_mac == 0)
				continue;

			ret = host_get_neigh_status(p->hwaddr);
			if (ret == 0 && has_ip) {
				/*Here the node is rechable so
				 * active status is 1
				 */
				dbg("ETH node is reachable " MACFMT "\n", MAC2STR(p->hwaddr));
				if (p->active != 1) {
					//Here as the node is reachable
					//we are adding in map-topology
					host_send_client_event(priv, p, 1);
					host_change_active_state(priv, p, 1);
				}
			} else {
				/*Here as the status is not reachable
				 */
				dbg("ETH node is unreachable " MACFMT "\n", MAC2STR(p->hwaddr));
				if (p->active != 0) {
					host_send_client_event(priv, p, 0);
					host_change_active_state(priv, p, 0);
				}
			}
		}
	}
}

void host_get_wifi_table_cb(struct topologyd_private *priv)
{
	struct host_node *p;
	int i = 0;
	char mac_str[18] = { 0 };

	dbg("Inside %s %d\n", __func__, __LINE__);
	if (!priv)
		return;

	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->host.node_htable[i]))
			continue;

		hlist_for_each_entry(p, &priv->host.node_htable[i], hlist) {
			//if (p->al_node)
			//	continue;
			if (hwaddr_is_zero(p->hwaddr))  // TODO: unlikely..
				continue;
			if (p->is_copy == 1)
				continue;

			hwaddr_ntoa(p->hwaddr, mac_str);
			dbg("macstr  is %s\n", mac_str);

			if (p->intf_type != HOST_TYPE_ETHER && !p->is_ieee1905_node && p->is_device_mac == 0) {
				if (p->active == 1) {
						time_t now;
					double diff_t;
					now = time(NULL);
					diff_t = difftime(now, p->age_time);
					if (diff_t > WIFI_TIME_OUT)
					host_topo_node_del(priv, NULL, p->hwaddr, 1);
					dbg("Timeout diff is ===================%f\n", diff_t);
				}
			}
		}
	}
}

void host_arp_table_refresh(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
		container_of(t, struct topologyd_private, host_arp_timer);

	dbg("Inside %s %d\n", __func__, __LINE__);

	host_get_table_cb(priv);
}

void host_wifi_table_refresh(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
		container_of(t, struct topologyd_private, host_wifi_timer);

	dbg("Inside %s %d\n", __func__, __LINE__);

	host_get_wifi_table_cb(priv);
}

int host_send_arp_req(char *ipaddr_str, struct host_config *cfg)
{
	char cmd[256] = {0};
	int i = 0;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (ipaddr_str == NULL)
		return -1;

	for (i = 0; i < cfg->num_ifname; i++) {
		snprintf(cmd, 256, ARPING_CMD, cfg->ifname[i], ipaddr_str);
		dbg("Inside %s %d cmd is %s\n", __func__, __LINE__, cmd);
		runCmd(cmd);
	}
	return 0;
}

/*
 *{ "topology.host": {"action":"connect","macaddr":"fe:ce:90:d7:5e:64","ipaddr":"192.168.1.157","network":"lan"} }
 *{ "topology.host": {"action":"disconnect","macaddr":"fe:ce:90:d7:5e:64","ipaddr":"192.168.1.157","network":"lan"} }
 */

void host_send_client_event(struct topologyd_private *priv, struct host_node *p, uint8_t state)
{
	char mac[18] = {0};
	struct blob_buf bb = {0};

	if (!p || !priv)
		return;

	dbg("Inside %s...\n", __func__);

	if (p->ignore == 1)
		return;

	blob_buf_init(&bb, 0);
	hwaddr_ntoa(p->hwaddr, mac);
	blobmsg_add_u8(&bb, "active", state);
	blobmsg_add_string(&bb, "macaddr", mac);
	blobmsg_add_string(&bb, "ipaddr", p->ipv4addr);
	blobmsg_add_string(&bb, "network", p->network);

	ubus_send_event(priv->ctx, "topology.host", bb.head);

	blob_buf_free(&bb);
}

void host_change_active_state(struct topologyd_private *priv, struct host_node *p, uint8_t state)
{
	char mac_str[18] = { 0 };
	time_t tmp_t;
	struct tm *info;
	char str_tm[32] = { 0 };

	dbg("Inside %s %d\n", __func__, __LINE__);
	hwaddr_ntoa(p->hwaddr, mac_str);
	p->active = state;
	p->active_last_change = time(NULL);
	//here we need to write this info in the config file
	config_add_default_host_mac("hosts", "host", mac_str, priv->host_global_cfg.persistent);
	if (state == 0x0)
		config_set_host_option("hosts", "host", "macaddr", mac_str, "active", "0", priv->host_global_cfg.persistent);
	else
		config_set_host_option("hosts", "host", "macaddr", mac_str, "active", "1", priv->host_global_cfg.persistent);

	tmp_t = p->active_last_change;
	info = gmtime(&tmp_t);
	strftime(str_tm, sizeof(str_tm), "%Y-%m-%dT%H:%M:%SZ", info);
	config_set_host_option("hosts", "host", "macaddr", mac_str, "active_last_change", str_tm, priv->host_global_cfg.persistent);
}

void host_remove_aged_node(struct topologyd_private *priv)
{
	int i = 0;
	char mac_str[18] = { 0 };
	struct host_node *p = NULL;
	struct hlist_node *tmp = NULL;

	if (priv == NULL)
		return;

	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->host.node_htable[i]))
			continue;

		hlist_for_each_entry_safe(p, tmp, &priv->host.node_htable[i], hlist) {
			//Here we take the current time and the
			//time the node has been inactive to check if it is
			//greater then the age if yes then node is deleted
			if (p->active == 0) {
				time_t now;
				double diff_t;
				now = time(NULL);
				diff_t = difftime(now, p->active_last_change);
				time_t limit = priv->host_global_cfg.age_timer * 60;
				if (diff_t > limit) {
					hwaddr_ntoa(p->hwaddr, mac_str);
					config_del_section("hosts", "host", mac_str, priv->host_global_cfg.persistent);
					dbg("Ageing timeout delecting mac node = ["MACFMT"]\n", MAC2STR(p->hwaddr));
					host_node_del(priv->host.node_htable, p->hwaddr, priv);
					priv->host.num_nodes--;
				}
			}
		}
	}
}

void host_set_interface_type(struct host_node *p, struct topologyd_private *priv)
{
	char mac_str[18] = { 0x0 };

	if (p == NULL || priv == NULL)
		return;

	hwaddr_ntoa(p->hwaddr, mac_str);
	config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type",
		(p->intf_type == HOST_TYPE_WIFI ? "wifi" :  "eth"), priv->host_global_cfg.persistent);
	config_set_host_option("hosts", "host", "macaddr", mac_str, "active", "1", priv->host_global_cfg.persistent);
}


#if 0
static void host_mid_response_cb(struct ubus_request *req,
		int mtype, struct blob_attr *msg)
{
	json_object *jroot = NULL;
	char *str = NULL;
	uint32_t msg_mid = 0;

	dbg("Inside %s %d\n", __func__, __LINE__);
	if (!msg)
		return;

	str = blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jroot = json_tokener_parse(str);
		free(str);
	}

	if (jroot) {
		msg_mid = _json_obj_get_int(jroot, "mid");
		json_safe_put(jroot);
	}
	dbg("Inside %s %d mid = [%d]\n", __func__, __LINE__, msg_mid);

	return;
}
#endif

void host_wifi_sta_event_handler(void *c, struct blob_attr *msg)
{

	struct topologyd_private *p = (struct topologyd_private *)c;
	char ifname[16] = {0}, event[16] = {0};
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	bool add = false, del = false;

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2])
		return;

	strncpy(ifname, blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	add = !strcmp(event, "connected");
	del = !strcmp(event, "disconnected");

	if (add || del) {
		struct blob_attr *data[1];
		static const struct blobmsg_policy data_attr[1] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		};

		char mac_str[18] = {0};
		uint8_t mac[6] = {0};

		blobmsg_parse(data_attr, 1, data, blobmsg_data(tb[2]),
			blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		strncpy(mac_str, blobmsg_data(data[0]), sizeof(mac_str) - 1);

		if (!hwaddr_aton(mac_str, mac))
			return;

		if (add) {
			struct host_node *hn;

			hn = host_node_lookup(p->host.node_htable, mac);
			if (!hn) {
				hn = host_topo_node_add(p, NULL, mac, 1);
				if (hn->is_ipaddr)
					host_send_client_event(p, hn, 1);
			} else {
				if (hn->active == 0) {
					hn = host_topo_node_add(p, NULL, mac, 1);
					if (hn->is_ipaddr)
						host_send_client_event(p, hn, 1);
				}
			}
		} else if (del)
			host_topo_node_del(p, NULL, mac, 1);
	}
}

void host_event_handler(struct ubus_context *ctx,
		struct ubus_event_handler *ev, const char *type,
		struct blob_attr *msg)
{
	char *str;

	struct topologyd_private *priv =
		container_of(ev, struct topologyd_private, wifi_evh);

	if (!msg)
		return;

	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	info("[ &Host = %p ] Received [event = %s]  [val = %s]\n",
		priv, type, str);
	host_wifi_sta_event_handler(priv, msg);
	free(str);
}

bool is_local_mapagent_available(void)
{
	bool enable = false;
	struct stat sb;
	int ret = 0;

	dbg("Inside %s ===>\n", __func__);
	ret = stat("/etc/config/mapagent", &sb);
	if (ret)
		return false;

	ret = host_config_mapagent_get(&enable);
	if (ret == 0) {
		if (enable == true)
			return true;
	}
	return false;
}

bool host_create_file(void)
{
	if (!access("/var/state/hosts", F_OK|R_OK)) /* flawfinder: ignore */
		return true;
	else {
		FILE *fptr = NULL;
		fptr = fopen("/var/state/hosts", "w+");

		if (fptr == NULL) {
			err("Unable to create file\n");
			return false;
		}
		fclose(fptr);
	}
	return true;
}


int host_config_defaults(struct host_config *cfg)
{
	if (!cfg)
		return -1;

	cfg->persistent = false;
	cfg->age_timer = HOST_AGE_TIMER_DEFAULT;
	strncpy((char *)cfg->ifname[0], "br-lan", 16);
	cfg->num_ifname = 1;

	return 0;
}

int host_copy_node_data(struct host_node *src_host,  struct host_node *dest_host,
		 struct topologyd_private *priv)
{
	int i = 0, n = 0, m = 0, k = 0;
	int found = 0;
	char mac_str[18] = { 0 };
	time_t tmp_t;
	struct tm *info;
	char str_tm[32] = { 0 };

	dbg("Inside %s %d\n", __func__, __LINE__);

	hwaddr_ntoa(dest_host->hwaddr, mac_str);
	strncpy(dest_host->device, src_host->device, 16);
	strncpy(dest_host->network, src_host->network, 16);
	dest_host->active = src_host->active;
	dest_host->active_last_change = src_host->active_last_change;
	//write this to the file
	if (src_host->active == 0x0)
		config_set_host_option("hosts", "host", "macaddr", mac_str, "active", "0", priv->host_global_cfg.persistent);
	else
		config_set_host_option("hosts", "host", "macaddr", mac_str, "active", "1", priv->host_global_cfg.persistent);

	tmp_t = src_host->active_last_change;
	info = gmtime(&tmp_t);
	strftime(str_tm, sizeof(str_tm), "%Y-%m-%dT%H:%M:%SZ", info);
	config_set_host_option("hosts", "host", "macaddr", mac_str, "active_last_change", str_tm, priv->host_global_cfg.persistent);

	strncpy(dest_host->ipv4addr, src_host->ipv4addr, 24);
	strncpy(dest_host->hostname, src_host->hostname, 64);
	dest_host->addrsrc = src_host->addrsrc;
	dest_host->leasetmrmn = src_host->leasetmrmn;
	dest_host->active_connections = src_host->active_connections;
	//Here we need to copy as in case of ethernet we need stats of br-lan
	if (dest_host->is_copy_lan == 1)
		memcpy(&dest_host->stats, &src_host->stats, sizeof(dest_host->stats));
	m = dest_host->ipv4addr_count;
	n = dest_host->ipv6addr_count;
	for (i = 0; i < src_host->ipv4addr_count; i++) {
		found = 0;
		for (k = 0; k < m; k++) {
			int ret1 = 0;
			ret1 = strncmp(dest_host->ipv4addr_list[k], "", 24);
			if (ret1 == 0) {
				m--;
				break;
			}
			ret1 = strncmp(dest_host->ipv4addr_list[k], src_host->ipv4addr_list[i], 24);
			if (ret1 == 0)
				found = 1;
		}
		if (found == 0) {
			strncpy(dest_host->ipv4addr_list[m], src_host->ipv4addr_list[i], 24);
			m++;
		}
	}

	for (i = 0; i < src_host->ipv6addr_count; i++) {
		found = 0;
		for (k = 0; k < n; k++) {
			int ret1 = 0;
			ret1 = strncmp(dest_host->ipv6addr[k], "", 128);
			if (ret1 == 0) {
				n--;
				break;
			}
			ret1 = strncmp(dest_host->ipv6addr[k], src_host->ipv6addr[i], 128);
			if (ret1 == 0)
				found = 1;
		}
		if (found == 0) {
			strncpy(dest_host->ipv6addr[n], src_host->ipv6addr[i], 128);
			n++;
		}
	}
	dest_host->ipv4addr_count = m;
	dest_host->ipv6addr_count = n;

	return 0;
}

/* prepare list of interface ip addresses in device
	list is used in host wan stats collection to skip conntrack entry
	created for communication between host and device(non wan traffic)*/
void host_prepare_device_address_list(struct topologyd_private *priv, int family)
{
	struct json_object *interfaces = NULL;
	int i = 0, len = 0;
	bool ret = 0;
	struct device_address_list *n;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!priv)
		return;

	if (!json_object_is_type(priv->network_dump, json_type_object))
		return;

	ret = json_object_object_get_ex(priv->network_dump, "interface", &interfaces);
	if (!ret || !interfaces)
		return;

	if (!json_object_is_type(interfaces, json_type_array))
		return;

	len = json_object_array_length(interfaces);

	for (i = 0; i < len; i++) {
		struct json_object *sta_obj = NULL, *jobj_network = NULL, *ipv4_arr = NULL;
		struct json_object *ipv6_arr = NULL;
		const char *network;
		int ipv4_len, j, ipv6_len;

		sta_obj = json_object_array_get_idx(interfaces, i);
		ret = json_object_object_get_ex(sta_obj, "interface", &jobj_network);
		if (!ret || !jobj_network)
			continue;

		network = json_object_get_string(jobj_network);
		if (!network)
			continue;

		if (family == AF_INET) {
			ret = json_object_object_get_ex(sta_obj, "ipv4-address", &ipv4_arr);
			if (!ret || !ipv4_arr)
				continue;

			if (!json_object_is_type(ipv4_arr, json_type_array))
				return;

			ipv4_len = json_object_array_length(ipv4_arr);
			for (j = 0; j < ipv4_len; j++) {
				struct json_object *jobj_ptr = NULL, *iface = NULL;
				const char *ipaddr;

				iface = json_object_array_get_idx(ipv4_arr, j);
				if (!iface)
					continue;

				ret = json_object_object_get_ex(iface, "address", &jobj_ptr);
				if (!ret || !jobj_ptr)
					continue;

				ipaddr = json_object_get_string(jobj_ptr);
				if (!ipaddr)
					continue;
				//allocated memory free is done in function host_wan_stats_update
				n = (struct device_address_list *)calloc(1, sizeof(struct device_address_list));
				if (!n)
					return;
				if (inet_pton(AF_INET, ipaddr, &(n->addr.ipv4_sin_addr.s_addr)) > 0) {
					n->addr.ipv4_sin_addr.s_addr = ntohl(n->addr.ipv4_sin_addr.s_addr);
					n->addr.family = AF_INET;
					hlist_add_head(&n->hlist, &priv->device_addr_list_table);
				}
				return;
			}
		}
		else if (family == AF_INET6) {
			ret = json_object_object_get_ex(sta_obj, "ipv6-prefix-assignment", &ipv6_arr);
			if (!ret || !ipv6_arr)
				continue;

			if (!json_object_is_type(ipv6_arr, json_type_array))
				return;
			ipv6_len = json_object_array_length(ipv6_arr);
			for (j = 0; j < ipv6_len; j++) {
				struct json_object *jobj_ptr = NULL, *iface = NULL, *local_addr=NULL;
				const char *ipaddr;

				iface = json_object_array_get_idx(ipv6_arr, j);
				if (!iface)
					continue;
				ret = json_object_object_get_ex(iface, "local-address", &local_addr);
				if (!ret || !local_addr)
					continue;
				ret = json_object_object_get_ex(local_addr, "address", &jobj_ptr);
				if (!ret || !jobj_ptr)
					continue;
				ipaddr = json_object_get_string(jobj_ptr);
				if (!ipaddr)
					continue;
				//allocated memory free is done in function host_wan_stats_update
				n = (struct device_address_list *) calloc(1, sizeof(struct device_address_list));
				if (n) {
					if (inet_pton(AF_INET6, ipaddr, &(n->addr.ipv6_sin_addr.s6_addr)) > 0) {
						n->addr.family = AF_INET6;
						hlist_add_head(&n->hlist, &priv->device_addr_list_table);
					}
				}
			}

			ret = json_object_object_get_ex(sta_obj, "ipv6-address", &ipv6_arr);
			if (!ret || !ipv6_arr)
				continue;

			if (!json_object_is_type(ipv6_arr, json_type_array))
				return;
			ipv6_len = json_object_array_length(ipv6_arr);
			for (j = 0; j < ipv6_len; j++) {
				struct json_object *jobj_ptr = NULL, *iface = NULL;
				const char *ipaddr;

				iface = json_object_array_get_idx(ipv6_arr, j);
				if (!iface)
					continue;
				ret = json_object_object_get_ex(iface, "address", &jobj_ptr);
				if (!ret || !jobj_ptr)
					continue;
				ipaddr = json_object_get_string(jobj_ptr);
				if (!ipaddr)
					continue;
				//allocated memory free is done in function host_wan_stats_update
				n = (struct device_address_list *) calloc(1, sizeof(struct device_address_list));
				if (!n)
					return;
				if (inet_pton(AF_INET6, ipaddr, &(n->addr.ipv6_sin_addr.s6_addr)) > 0) {
					n->addr.family = AF_INET6;
					hlist_add_head(&n->hlist, &priv->device_addr_list_table);
				}
			}
		}
	}
	return;
}

/*Here the below are checked for each host node in case of periodic run
 * 1. If Ipaddress is fetched from dhcp lease
 * 2. If node is eth then check reachable
 * 3. If node is eth and not reachable do arping
 */
void host_run_status_check(struct topologyd_private *priv)
{
	struct host_node *p;
	int i = 0, ret = -1;
	char mac_str[18] = { 0 };

	dbg("Inside %s %d\n", __func__, __LINE__);
	if (!priv)
		return;

	host_get_interface_type(priv);
	host_ieee1905_call(priv);

	//host_remove_aged_node(priv);
	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->host.node_htable[i]))
			continue;

		hlist_for_each_entry(p, &priv->host.node_htable[i], hlist) {
			bool has_ip = p->is_ipaddr;

			if (hwaddr_is_zero(p->hwaddr))  // TODO: unlikely..
				continue;
			if (p->is_copy == 1)
				continue;

			hwaddr_ntoa(p->hwaddr, mac_str);
			host_get_wifi_data(priv, p);

			if (p->intf_type != HOST_TYPE_ETHER && !p->is_ieee1905_node && p->is_device_mac == 0) {
				//memset(p->device, 0, 16);
				dbg("%s %d hasip = %d, %s\n", __func__, __LINE__, has_ip, p->ipv4addr);

				ret = host_get_ipaddr_hostname(priv, mac_str, p);
				if (ret != 0) {
					err("Unable to fetch ip adddress and host name\n");
					p->is_ipaddr = 0;
				} else {
					p->is_ipaddr = 1;
					if (!has_ip) {
						if (p->is_in_assoclist) {
							host_send_client_event(priv, p, 1);
						} else if (p->intf_type == HOST_TYPE_WIFI && !p->is_in_assoclist) {
							if (p->active == 1)
								host_send_client_event(priv, p, 1);
						}
					}
				}
				dbg("%s %d hasip = %d, %s\n", __func__, __LINE__, has_ip, p->ipv4addr);
				continue;
			} else if (p->is_ieee1905_node)
				memset(p->device, 0, 16);

			if (host_get_ipaddr_hostname(priv, mac_str, p))
				p->is_ipaddr = 0;
			else
				p->is_ipaddr = 1;
			char ipv4addr[24] = { 0 };

			if (strncmp(p->ipv4addr, ipv4addr, 24)) {
				ret = host_send_arp_req(p->ipv4addr, &priv->host_global_cfg);
			} else {
				dbg("ETH node is unreachable " MACFMT "\n", MAC2STR(p->hwaddr));
				if (p->active != 0) {
					host_send_client_event(priv, p, 0);
					host_change_active_state(priv, p, 0);
				}
			}
		}
	}
	/* Here we set the timer to check the arp request
	*/
	uloop_timeout_set(&priv->host_arp_timer, HOST_ARP_TIMER * 1000);
	if (priv->is_mapagent)
		uloop_timeout_set(&priv->host_wifi_timer, HOST_WIFI_TIMER * 1000);

}

/*Here the host periodic refresh timer does as below:
 * 1. run the topology refresh algo such that all the new
 *    non -1905 neighbors can be fetched
 * 2. Check all the nodes of type ETHER and get there status
 *    from the arp table.
 */
void host_periodic_refresh(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
		container_of(t, struct topologyd_private, host_refresh_timer);

	dbg("Refresh host topology...\n");

	/*here we send the topology query to selfnode each*/
	/*time the algo runs to keep self node updated */
	topologyd_send_ieee1905_topology_query(priv, NULL);
	host_run_status_check(priv);
	uloop_timeout_set(&priv->host_refresh_timer, HOST_REFRESH_TIMER * 1000);
}
